from .fields import FieldData


class RegsDotGovData(FieldData):
    """Class for processing docket data sourced from Regulations.gov.
    Inherits from `FieldData`.
    """
    def __init__(self, 
                 documents: list[dict], 
                 field_key: str = "regulations_dot_gov_info",
                 subfield_key: str = "docket_id", 
                 value_key: str = None
                 ) -> None:
        super().__init__(documents=documents, field_key=field_key, subfield_key=subfield_key)
        if value_key is None:
            self.value_key = subfield_key
        else:
            self.value_key = value_key

    def _extract_field_info(self, document: dict) -> str | None:
        
        field_info = document.get(self.field_key, {})
        
        if len(field_info) == 0:
            values = None
        elif isinstance(field_info, list):
            values = [docket.get(self.subfield_key) for docket in field_info]
        elif isinstance(field_info, dict):
            values = [field_info.get(self.subfield_key)]

        if values is not None:
            #print(values)
            try_alt = all(("FRDOC" in v) for v in values if v is not None)
            if try_alt:
                values = self._try_alt_key(document)
            else:
                values = "; ".join(v for v in values if v is not None)
        
        return values
    
    def _try_alt_key(self, document: dict, alt_key: str = "docket_ids") -> str | None:
        
        field_info = document.get(alt_key, {})
        
        if len(field_info) == 0:
            values = None
        elif isinstance(field_info, list):
            values = field_info
        
        if values is not None:
            values = "; ".join(v for v in values if "FRDOC" not in v)
        
        return values


class Dockets(RegsDotGovData):
    """Class for processing docket data from "dockets" field.
    Inherits from `RegsDotGovData`.
    """
    def __init__(self, 
                 documents: list[dict], 
                 field_key: str = "dockets", 
                 subfield_key: str = "id", 
                 value_key: str = "docket_id"
                 ) -> None:
        super().__init__(documents=documents, field_key=field_key, subfield_key=subfield_key, value_key=value_key)



from itertools import product
import json
from pathlib import Path

from requests import get

from fr_toolbelt.preprocessing import (
    AgencyMetadata, 
    AgencyData, )


# TEST OBJECTS AND UTILS #


TESTS_PATH = Path(__file__).parent

ENDPOINT_URL = r"https://www.federalregister.gov/api/v1/documents.json?"

TEST_PARAMS_FULL = {
    "per_page": 1000, 
    "page": 0, 
    "order": "oldest", 
    "conditions[publication_date][gte]": "2023-11-01", 
    "conditions[publication_date][lte]": "2023-11-30"
    }
TEST_RESPONSE_FULL = get(ENDPOINT_URL, TEST_PARAMS_FULL).json()

TEST_PARAMS_PARTIAL = {
    "per_page": 1000, 
    "page": 0, 
    "order": "oldest", 
    "conditions[publication_date][gte]": "2023-01-01", 
    "conditions[publication_date][lte]": "2023-06-30"
    }
TEST_URL_PARTIAL = get(ENDPOINT_URL, TEST_PARAMS_PARTIAL).url
TEST_RESPONSE_PARTIAL = get(ENDPOINT_URL, TEST_PARAMS_PARTIAL).json()
TEST_COUNT_PARTIAL = TEST_RESPONSE_PARTIAL["count"]

with open(TESTS_PATH / "test_documents.json", "r", encoding="utf-8") as f:
    TEST_DATA = json.load(f).get("results", [])

TEST_METADATA, TEST_SCHEMA = AgencyMetadata().get_agency_metadata()


# preprocessing.agencies #


def test_agencies_metadata_init():
    agency_metadata = AgencyMetadata()
    assert isinstance(agency_metadata.data, list)
    assert len(agency_metadata.transformed_data) > 0
    assert isinstance(agency_metadata.transformed_data, dict)
    assert len(agency_metadata.schema) > 0
    assert isinstance(agency_metadata.schema, list)


def test_agencies_metadata_get():
    agency_metadata = AgencyMetadata()
    metadata, schema = agency_metadata.get_agency_metadata()
    assert isinstance(metadata, dict)
    assert isinstance(schema, list)


def test_agencies_metadata_save_metadata(path = TESTS_PATH, file_name: str = "test_metadata.json"):
    test_path = path / file_name
    assert not test_path.exists()
    agency_metadata = AgencyMetadata()
    agency_metadata.save_metadata(path, file_name=file_name)
    assert test_path.is_file()
    if test_path.is_file():
        test_path.unlink()


def test_agencies_metadata_save_schema(path = TESTS_PATH, file_name: str = "test_schema.json"):
    test_path = path / file_name
    assert not test_path.exists()
    agency_metadata = AgencyMetadata()
    agency_metadata.save_schema(path, file_name=file_name)
    assert test_path.is_file()
    if test_path.is_file():
        test_path.unlink()


def test_agencies_data_init(
        documents = TEST_DATA, 
        metadata = TEST_METADATA, 
        schema = TEST_SCHEMA
    ):
    agency_data = AgencyData(documents=documents, metadata=metadata, schema=schema)
    assert isinstance(agency_data.documents, list)
    assert len(agency_data.documents) > 0
    assert isinstance(agency_data.metadata, dict)
    assert len(agency_data.metadata) > 0
    assert isinstance(agency_data.schema, dict), f"schema should be type dict but it is type {type(agency_data.schema)}"
    assert len(agency_data.schema) == 3


def test_agencies_data_process_defaults(
        documents = TEST_DATA, 
        metadata = TEST_METADATA, 
        schema = TEST_SCHEMA, 
        check_keys = ("agency_slugs", "parent_slug", "subagency_slug", "independent_reg_agency")
):
    agency_data = AgencyData(documents=documents, metadata=metadata, schema=schema)
    processed = agency_data.process_data()
    assert isinstance(processed, list)
    assert len(processed) > 0
    processed_keys = set((k for d in processed for k in d))
    for key in check_keys:
        assert key in processed_keys, f"Output missing {key=}"
    for key in agency_data.field_keys:
        assert key not in processed_keys, f"Failed to delete keys: {key=}"


def test_agencies_data_process_return_format_str(
        documents = TEST_DATA, 
        metadata = TEST_METADATA, 
        schema = TEST_SCHEMA, 
        check_keys = ("agency_slugs", "independent_reg_agency", ), 
        check_prefixes = ("parent_", "subagency_", ), 
        return_format="short_name", 
):
    agency_data = AgencyData(documents=documents, metadata=metadata, schema=schema)
    processed = agency_data.process_data(return_format=return_format)
    assert isinstance(processed, list)
    assert len(processed) > 0
    processed_keys = set((k for d in processed for k in d))
    checks = list(check_keys) + [f"{p[0]}{p[1]}" for p in product(check_prefixes, (return_format, ))]
    for key in checks:
        assert key in processed_keys, f"Output missing {key=}"
    for key in agency_data.field_keys:
        assert key not in processed_keys, f"Failed to delete keys: {key=}"


def test_agencies_data_process_return_format_tuple(
        documents = TEST_DATA, 
        metadata = TEST_METADATA, 
        schema = TEST_SCHEMA, 
        check_keys = ("agency_slugs", "independent_reg_agency", ), 
        check_prefixes = ("parent_", "subagency_", ), 
        return_format=("short_name", "slug", ), 
):
    agency_data = AgencyData(documents=documents, metadata=metadata, schema=schema)
    processed = agency_data.process_data(return_format=return_format)
    assert isinstance(processed, list)
    assert len(processed) > 0
    processed_keys = set((k for d in processed for k in d))
    checks = list(check_keys) + [f"{p[0]}{p[1]}" for p in product(check_prefixes, return_format)]
    for key in checks:
        assert key in processed_keys, f"Output missing {key=}"
    for key in agency_data.field_keys:
        assert key not in processed_keys, f"Failed to delete keys: {key=}"



from .agencies import AgencyMetadata, AgencyData
from .dockets import RegsDotGovData, Dockets
from .presidents import Presidents
from .rin import RegInfoData


class PreprocessingError:
    pass


def process_documents(
        documents: list[dict], 
        which: str | list | tuple = "all", 
        docket_data_source: str = "dockets", 
        del_keys: str | list | tuple = None, 
        **kwargs
    ) -> list[dict]:
    """Process one or more fields in each document.

    Args:
        documents (list[dict]): Documents to process.
        which (str | list | tuple, optional): Which fields to process per document. Defaults to "all". Valid inputs include "all" or some combination of "agencies", "dockets", "presidents", "rin".
        docket_data_source (str, optional): Select which field to use as a source for processing dockets data. Defaults to "dockets". Valid inputs include "regulations_dot_gov_info" and "dockets".
        del_keys (str | list | tuple, optional): Delete select keys from results. Defaults to None.

    Raises:
        PreprocessingError: Failed to preprocess input documents.

    Returns:
        list[dict]: Processed documents.
    """
    # dictionary of alternative sources
    source_dict = {
        "dockets": Dockets, 
        "regulations_dot_gov_info": RegsDotGovData
        }
    
    # maps keyword to class
    process_fields = {
        "agencies": AgencyData, 
        "dockets": source_dict.get(docket_data_source, Dockets), 
        "presidents": Presidents, 
        "rin": RegInfoData, 
        }
    
    # process documents
    if (which == "all") or ("all" in which and isinstance(which, (list, tuple))):
    
        for field, function in process_fields.items():
            if field == "agencies":
                metadata, schema = AgencyMetadata().get_agency_metadata()
                documents = function(documents, metadata, schema).process_data(**kwargs)
            else:
                documents = function(documents).process_data()
    
    elif isinstance(which, str) and (which in process_fields.keys()):
        if which == "agencies":
            metadata, schema = AgencyMetadata().get_agency_metadata()
            documents = process_fields[which](documents, metadata, schema).process_data(**kwargs)
        else:
            documents = process_fields[which](documents).process_data()
    
    elif isinstance(which, (list, tuple)):
        valid_fields = (w for w in which if w in process_fields.keys())
        for field in valid_fields:
            if field == "agencies":
                metadata, schema = AgencyMetadata().get_agency_metadata()
                documents = process_fields[field](documents, metadata, schema).process_data(**kwargs)
            else:
                documents = process_fields[field](documents).process_data()
    
    else:
        raise PreprocessingError("Failed to preprocess input documents.")
    
    # delete keys if passed
    if del_keys is not None:
        return [{k: v for k, v in doc.items() if ((k != del_keys) and (k not in del_keys))} for doc in documents]
    else:
        return documents



from datetime import date
from fr_toolbelt.utils import DateFormatter


def test__convert_to_datetime_date(
        success = ("2024-01-01", "20240101", "2024-W01-1", date(2024, 1, 1))
    ):
    
    for attempt in success:
        fdate = DateFormatter(attempt)
        result = fdate._DateFormatter__convert_to_datetime_date(attempt)
        assert isinstance(result, date)


def test_get_year_self(
        input_success: dict = {"string": "2023-01-01", "year": 2023}, 
        input_fail: str = "01/01/2023"
    ):
    
    fdate = DateFormatter(input_success.get("string"))
    year = fdate.get_year()
    assert isinstance(year, int)
    assert year == input_success.get("year")
    
    try:
        DateFormatter(input_fail).get_year()
    except ValueError as e:
        assert e.__class__ == ValueError


def test_get_year_alt(
        input_success: dict = {
            1: {"string": "2023-01-01", "year": 2023}, 
            2: {"string": "2024-01-01", "year": 2024}
            }
    ):
    
    fdate = DateFormatter(input_success.get(1).get("string"))
    year = fdate.get_year(input_success.get(2).get("string"))
    assert isinstance(year, int)
    assert year == input_success.get(2).get("year")


def test_get_formatted_date(
        success = ("2024-01-01", "20240101", "2024-W01-1", date(2024, 1, 1))
    ):
    for attempt in success:
        result = DateFormatter(attempt).get_formatted_date()
        assert isinstance(result, date)


def test_date_in_quarter():
    
    attempt1, attempt2, year, quarter = date(2023, 1, 1), date(2023, 4, 1), "2023", "Q1"
    
    fdate1 = DateFormatter(attempt1)
    result = fdate1.date_in_quarter(year, quarter)
    assert attempt1 == result
    
    fdate2 = DateFormatter(attempt2)
    result = fdate2.date_in_quarter(year, quarter)
    assert attempt2 != result
    assert f"{result}" == f"{year}-{fdate2.quarter_schema.get(quarter)[1]}", "should return end of Q1"
    
    result = fdate2.date_in_quarter(year, quarter, return_quarter_end=False)
    assert attempt2 != result
    assert f"{result}" == f"{year}-{fdate2.quarter_schema.get(quarter)[0]}", "should return beginning of Q1"


def test_gt_date_exclusive(inputs = {
        "earlier": date(2023, 1, 1), 
        "later": date(2023, 4, 1)}
    ):
    fdate = DateFormatter(inputs.get("later"))
    res = fdate.greater_than_date(inputs.get("earlier"))
    assert res, "later > earlier"
    res2 = fdate.greater_than_date(inputs.get("later"))
    assert not res2, "later !> later (exclusive)"
    
    fdate = DateFormatter(inputs.get("earlier"))
    res = fdate.greater_than_date(inputs.get("later"))
    assert not res, "earlier !> later"
    res2 = fdate.greater_than_date(inputs.get("earlier"))
    assert not res2, "earlier !> earlier (exclusive)"


def test_gt_date_inclusive(inputs = {
        "earlier": date(2023, 1, 1), 
        "later": date(2023, 4, 1)
        }
    ):
    fdate = DateFormatter(inputs.get("later"))
    res = fdate.greater_than_date(inputs.get("later"), inclusive=True)
    assert res, "later == later (inclusive)"
    
    fdate = DateFormatter(inputs.get("earlier"))
    res = fdate.greater_than_date(inputs.get("earlier"), inclusive=True)
    assert res, "earlier == earlier (inclusive)"


def test_lt_date_exclusive(inputs = {
        "earlier": date(2023, 1, 1), 
        "later": date(2023, 4, 1)
        }
    ):
    fdate = DateFormatter(inputs.get("later"))
    res = fdate.less_than_date(inputs.get("earlier"))
    assert not res, "later !< earlier"
    res2 = fdate.less_than_date(inputs.get("later"))
    assert not res2, "later !< later (exclusive)"
    
    fdate = DateFormatter(inputs.get("earlier"))
    res = fdate.less_than_date(inputs.get("later"))
    assert res, "earlier < later"
    res2 = fdate.less_than_date(inputs.get("earlier"))
    assert not res2, "earlier !< earlier (exclusive)"


def test_lt_date_inclusive(inputs = {
        "earlier": date(2023, 1, 1), 
        "later": date(2023, 4, 1)
        }
    ):
    fdate = DateFormatter(inputs.get("later"))
    res = fdate.less_than_date(inputs.get("later"), inclusive=True)
    assert res, "later == later (inclusive)"
    
    fdate = DateFormatter(inputs.get("earlier"))
    res = fdate.less_than_date(inputs.get("earlier"), inclusive=True)
    assert res, "earlier == earlier (inclusive)"



from .fields import FieldData


class Presidents(FieldData):
    """Class for processing president data.
    Inherits from `FieldData`.
    """    
    def __init__(
            self, 
            documents: list[dict], 
            field_key: str = "president", 
            subfield_key: str = "identifier", 
            value_key: str = "president_id", 
            schema: dict = {
                "transition_years": (1993, 2001, 2009, 2017, 2021), 
                "presidents": ("william-j-clinton", "george-w-bush", "barack-obama", "donald-trump", "joe-biden"), 
                }
        ) -> None:
        super().__init__(documents=documents, field_key=field_key, subfield_key=subfield_key, value_key=value_key)
        self.schema = schema

    def _extract_field_info(self, document: dict) -> str | None:
        
        field_info = document.get(self.field_key, {})
        
        if len(field_info) == 0:
            values = None
        else:
            values = field_info.get(self.subfield_key)

        return values



from datetime import datetime, date, timedelta
import json
from pathlib import Path

from requests import get

from fr_toolbelt.api_requests import (
    _retrieve_results_by_next_page, 
    get_documents_by_date, 
    get_documents_by_number, 
    )


# TEST OBJECTS AND UTILS #


TESTS_PATH = Path(__file__).parent

ENDPOINT_URL = r"https://www.federalregister.gov/api/v1/documents.json?"

TEST_PARAMS_FULL = {
    "per_page": 1000, 
    "page": 0, 
    "order": "oldest", 
    "conditions[publication_date][gte]": "2023-11-01", 
    "conditions[publication_date][lte]": "2023-11-30"
    }
TEST_RESPONSE_FULL = get(ENDPOINT_URL, TEST_PARAMS_FULL).json()

TEST_PARAMS_PARTIAL = {
    "per_page": 1000, 
    "page": 0, 
    "order": "oldest", 
    "conditions[publication_date][gte]": "2023-01-01", 
    "conditions[publication_date][lte]": "2023-06-30"
    }
TEST_URL_PARTIAL = get(ENDPOINT_URL, TEST_PARAMS_PARTIAL).url
TEST_RESPONSE_PARTIAL = get(ENDPOINT_URL, TEST_PARAMS_PARTIAL).json()
TEST_COUNT_PARTIAL = TEST_RESPONSE_PARTIAL["count"]

with open(TESTS_PATH / "test_documents.json", "r", encoding="utf-8") as f:
    TEST_DATA = json.load(f).get("results", [])


# api_requests.get_documents #


def test_retrieve_results_by_next_page_full(
    endpoint_url: str = ENDPOINT_URL, 
    dict_params: dict = TEST_PARAMS_FULL, 
    test_response = TEST_RESPONSE_FULL
    ):
    
    results = _retrieve_results_by_next_page(endpoint_url, dict_params)
    assert len(results) == test_response.get("count")


def test_retrieve_results_by_next_page_partial(
    endpoint_url: str = ENDPOINT_URL, 
    dict_params: dict = TEST_PARAMS_PARTIAL
    ):
    
    results = _retrieve_results_by_next_page(endpoint_url, dict_params)
    assert len(results) == 10000, f"Should return 10,000; compare to API call: {TEST_URL_PARTIAL}"


def test_get_documents_by_date(start = "2024-01-01", end = "2024-01-31"):
    results, count = get_documents_by_date(start, end)
    assert isinstance(results, list)
    assert count == len(results)


def test_get_documents_by_date_quarters(start = "2022-12-01", end = "2023-02-01"):
    results, count = get_documents_by_date(start, end)
    assert isinstance(results, list)
    assert count == len(results)
    

def test_get_documents_by_date_types(
        start = "2024-01-01", 
        end = "2024-01-31", 
        types = ["RULE", "PRORULE"], 
        type_schema = {
            "RULE": "Rule", 
            "PRORULE": "Proposed Rule", 
            "NOTICE": "Notice", 
            "PRESDOCU": "Presidential Document"
            }
    ):
    results, count = get_documents_by_date(start, end, document_types=types)
    assert isinstance(results, list)
    assert count == len(results)
    res_types = set(doc.get("type") for doc in results)
    assert len(res_types) <= len(types)
    assert all(1 if r in (type_schema.get(t) for t in types) else 0 for r in res_types)


def test_get_documents_by_date_above_max_threshold(start = "2020-01-01", end = "2022-12-31", max = 10000):
    results, count = get_documents_by_date(start, end)
    assert count > max
    assert isinstance(results, list)
    assert count == len(results)


def test_get_documents_by_date_no_end_date(delta = 365):
    start = (datetime.now() - timedelta(delta)).date()
    test_error = "will remain string if error is not handled in try/except block"
    try:
        results, count = get_documents_by_date(start)
    except TypeError as err:
        test_error = err
        results, count = get_documents_by_date(start, end_date=date.today())
    max_date = max(date.fromisoformat(r.get("publication_date")) for r in results)
    assert isinstance(test_error, str), "Error was handled in try/except block; bug remains in program"
    assert max_date == date.today()
    assert isinstance(results, list)
    assert count == len(results)


def test_get_documents_by_number(numbers = ["2024-02204", "2023-28203", "2023-25797"]):
    results, count = get_documents_by_number(numbers)
    assert isinstance(results, list)
    assert count == len(results)



"""
Making requests from the Federal Register API.
"""

from .get_documents import (
    BASE_URL,
    BASE_PARAMS,
    DEFAULT_FIELDS, 
    QueryError,
    InputFileError, 
    get_documents_by_date, 
    get_documents_by_number, 
    parse_document_numbers, 
    _retrieve_results_by_next_page
)

__all__ = [
    "BASE_URL",
    "BASE_PARAMS",
    "DEFAULT_FIELDS",
    "QueryError",
    "InputFileError",
    "get_documents_by_date", 
    "get_documents_by_number", 
    "parse_document_numbers", 
    ]



from pathlib import Path
import csv

import pytest

from fr_toolbelt.api_requests import parse_document_numbers, InputFileError


TEST_PATH = Path(__file__).parent


def _create_test_csv_to_parse(path: Path):
    
    with open(path / 'test_parse_document_numbers.csv', 'w', newline='') as csvfile:
        fieldnames = ['document_number', 'html_url']
        writer = csv.DictWriter(csvfile, fieldnames=fieldnames)

        writer.writeheader()
        writer.writerow({'document_number': '2024-02204', 'html_url': 'https://federalregister.gov/d/2024-02204'})
        writer.writerow({'document_number': '2023-28203', 'html_url': 'https://federalregister.gov/d/2024-28203'})
        writer.writerow({'document_number': '2023-25797', 'html_url': 'https://federalregister.gov/d/2024-25797'})


@pytest.fixture(scope="session")
def get_csv_file(tmp_path_factory):
    fn = tmp_path_factory.mktemp("data")
    _create_test_csv_to_parse(path=fn)
    return fn


def test_parse_document_numbers_csv(get_csv_file):
    
    results = parse_document_numbers(get_csv_file)
    assert isinstance(results, list)
    assert len(results) == 3


def test_parse_document_numbers_error(path: Path = TEST_PATH):
    test_error = None
    results_out = None
    try:
        results_out = parse_document_numbers(path)
    except InputFileError as e:
        test_error = e
        assert isinstance(e, InputFileError)
    assert test_error is not None, f"{test_error=}"
    assert results_out is None



from datetime import date
import re


class DateFormatError(Exception):
    pass


class DateFormatter:
    
    def __init__(self, input_date: date | str) -> None:
        self.input_date = input_date
        self.formatted_date: date = self.__convert_to_datetime_date(input_date)
        self.year: int | None = self.formatted_date.year
        self.quarter_schema = {
            "Q1": ("01-01", "03-31"), 
            "Q2": ("04-01", "06-30"), 
            "Q3": ("07-01", "09-30"), 
            "Q4": ("10-01", "12-31"), 
            }
    
    def __extract_year(self, input_date: str | date):
        """Extract year from a string in a format similar to `datetime.datetime` or `datetime.date`.

        Args:
            alt_input_date (str | date): String of date.
        
        Returns:
            int: Year attribute of `datetime.date` object.
        """
        if isinstance(input_date, str):
            res = re.compile(r"\d{4}-\d{2}-\d{2}", re.I).match(input_date)
            if isinstance(res, re.Match):
                year = date.fromisoformat(res[0]).year
        elif isinstance(input_date, date):
            year = input_date.year
        else:
            year = None
        
        return year

    def __convert_to_datetime_date(self, input_date: date | str) -> date:
        """Converts `self.input_date` from `str` to `datetime.date`. Returns input if already in proper format.
        Input string should be in any valid ISO 8601 format (e.g., for Jan. 1, 2024 -> 2024-01-01, 20240101, 2024-W01-1).

        Raises:
            TypeError: Inappropriate argument type for input_date parameter.

        Returns:
            date: Date object.
        """    
        if isinstance(input_date, date):
            return input_date
        elif isinstance(input_date, str):
            return date.fromisoformat(input_date)
        else:
            raise TypeError(f"Inappropriate argument type {type(input_date)} for parameter 'input_date'.")

    def get_formatted_date(self) -> date:
        """Get `self.formatted_date` instance attribute.

        Returns:
            date: Formatted version of input date.
        """
        return self.formatted_date
    
    def get_year(self, alt_input_date: str | date = None) -> int:
        """Get `self.year` instance attribute

        Returns:
            int: Year of input date.
        """
        if alt_input_date is not None:
            return self.__extract_year(input_date=alt_input_date)
        else:
            return self.year
    
    def date_in_quarter(self, check_year: str, check_quarter: str, return_quarter_end: bool = True) -> date:
        """Checks if given date falls within a year's quarter. 
        Returns input date if True, otherwise returns first or last day of quarter.

        Args:
            check_year (str): Year to check against.
            check_quarter (str): Quarter to check against.
            return_quarter_end (bool, optional): Return end date of quarter when input not in range. Defaults to True.

        Raises:
            TypeError: Inappropriate argument type for 'input_date'.

        Returns:
            datetime.date: Returns input_date when it falls within range; otherwise returns boundary date of quarter.
        """
        quarter_range = self.quarter_schema.get(f"{check_quarter}".upper())
        check_date = self.__convert_to_datetime_date(self.input_date)
        range_start = date.fromisoformat(f"{check_year}-{quarter_range[0]}")
        range_end = date.fromisoformat(f"{check_year}-{quarter_range[1]}")
        in_range = (check_date >= range_start) and (check_date <= range_end)
        #return in_range
        if in_range:
            return check_date
        elif (not in_range) and return_quarter_end:
            return range_end
        elif (not in_range) and (not return_quarter_end):
            return range_start
        else:
            raise DateFormatError

    def greater_than_date(self, comparison_date: date | str, inclusive: bool = False) -> bool:
        """Compare whether a formatted date occurs after a given comparison date.

        Args:
            comparison_date (date | str): A given date to compare with the formatted date.

        Returns:
            bool: True if formatted date occurs after comparison date.
        """    
        if inclusive:
            return self.formatted_date >= self.__convert_to_datetime_date(comparison_date)
        else:
            return self.formatted_date > self.__convert_to_datetime_date(comparison_date)
    
    def less_than_date(self, comparison_date: date | str, inclusive: bool = False) -> bool:
        """Compare whether a formatted date occurs after a given comparison date.

        Args:
            comparison_date (date | str): A given date to compare with the formatted date.

        Returns:
            bool: True if formatted date occurs before comparison date.
        """
        if inclusive:
            return self.formatted_date <= self.__convert_to_datetime_date(comparison_date)
        else:
            return self.formatted_date < self.__convert_to_datetime_date(comparison_date)



"""
Utilities for other fr-toolbelt modules and beyond.
"""

from .duplicates import (
    identify_duplicates, 
    remove_duplicates, 
    flag_duplicates, 
    process_duplicates, 
    DuplicateError
    )

from .format_dates import DateFormatter, DateFormatError

__all__ = [
    "identify_duplicates", 
    "remove_duplicates",
    "flag_duplicates",
    "process_duplicates",
    "DuplicateError",
    "DateFormatter", 
    "DateFormatError", 
]




from abc import ABC, abstractmethod


class FieldData(ABC):
    """Base class for processing Federal Register fields."""    
    def __init__(self, 
                 documents: list[dict], 
                 field_key: str = None,
                 subfield_key: str = None, 
                 subfield_keys: tuple[str] = (),
                 value_key: str = None,
                 value_keys: tuple[str] = ()
                 ) -> None:
        self.documents = documents
        self.field_key = field_key
        self.subfield_key = subfield_key
        self.subfield_keys = subfield_keys
        self.value_key = value_key
        self.value_keys = value_keys
    
    @abstractmethod
    def _extract_field_info(self, document: dict):
        pass

    def _create_value_key(self, document: dict, values: str = None) -> dict:
        document_copy = document.copy() 
        document_copy.update(
            {self.value_key: values, }
            )
        return document_copy
    
    def _create_value_keys(self, document: dict, values: tuple = None) -> dict:

        document_copy = document.copy()
        # values: rin_info tuples (RIN, Priority, UA issue)
        if values is None:
            document_copy.update(
                {k: None for k in self.value_keys}
                )
        else:
            document_copy.update(
                {k: v for k, v in zip(self.value_keys, values)}
                )
        return document_copy
    
    def _del_field_key(self, document: dict, add_keys: str | tuple | list = None):
        document_copy = document.copy()
        if add_keys is not None:
            if isinstance(add_keys, str):
                keys = [add_keys, self.field_key]
            elif isinstance(add_keys, tuple | list):
                keys = list(add_keys) + [self.field_key]
            else:
                raise TypeError(f"Parameter add_keys must be `str`, `list`, or `tuple`; received {type(add_keys)}.")
            for key in keys:
                document_copy.pop(key, None)
        else:
            document_copy.pop(self.field_key, None)
        return document_copy
        
    def process_data(self, del_keys: str | tuple | list = None) -> list[dict]:
        
        if self.value_key is not None:
            return [self._del_field_key(self._create_value_key(doc, values=self._extract_field_info(doc)), add_keys=del_keys) for doc in self.documents]
        else:
            return [self._del_field_key(self._create_value_keys(doc, values=self._extract_field_info(doc)), add_keys=del_keys) for doc in self.documents]



"""
This patch fixes an issue where unwanted characters appeared after the progress bar in the terminal (see [GitHub issue 101](https://github.com/verigak/progress/issues/101)).
The code is taken from an [open pull request](https://github.com/verigak/progress/pull/109) by @james-duvall.
This script ensure that the patched code overwrites the imported package's Infinite class (inspired by [this solution](https://github.com/verigak/progress/issues/58#issuecomment-471718558) from @an0ndev).
"""

from __future__ import division, print_function

from collections import deque
from datetime import timedelta
from sys import stderr
import cursor
try:
    from time import monotonic
except ImportError:
    from time import time as monotonic


# source: https://github.com/verigak/progress/pull/109
class PatchedInfinite(object):
    """Patched by @james-duvall in pull request: https://github.com/verigak/progress/pull/109."""
    file = stderr
    sma_window = 10         # Simple Moving Average window
    check_tty = True
    hide_cursor = True

    def __init__(self, message='', **kwargs):
        self.index = 0
        self.start_ts = monotonic()
        self.avg = 0
        self._avg_update_ts = self.start_ts
        self._ts = self.start_ts
        self._xput = deque(maxlen=self.sma_window)
        for key, val in kwargs.items():
            setattr(self, key, val)

        self._max_width = 0
        self._hidden_cursor = False
        self.message = message

        if self.file and self.is_tty():
            if self.hide_cursor:
                cursor.hide()
                self._hidden_cursor = True
        self.writeln('')

    def __del__(self):
        if self._hidden_cursor:
            cursor.show()

    def __getitem__(self, key):
        if key.startswith('_'):
            return None
        return getattr(self, key, None)

    @property
    def elapsed(self):
        return int(monotonic() - self.start_ts)

    @property
    def elapsed_td(self):
        return timedelta(seconds=self.elapsed)

    def update_avg(self, n, dt):
        if n > 0:
            xput_len = len(self._xput)
            self._xput.append(dt / n)
            now = monotonic()
            # update when we're still filling _xput, then after every second
            if (xput_len < self.sma_window or
                    now - self._avg_update_ts > 1):
                self.avg = sum(self._xput) / len(self._xput)
                self._avg_update_ts = now

    def update(self):
        pass

    def start(self):
        pass

    def writeln(self, line):
        if self.file and self.is_tty():
            width = len(line)
            if width < self._max_width:
                # Add padding to cover previous contents
                line += ' ' * (self._max_width - width)
            else:
                self._max_width = width
            print('\r' + line, end='', file=self.file)
            self.file.flush()

    def finish(self):
        if self.file and self.is_tty():
            print(flush = True, file=self.file)
            if self._hidden_cursor:
                cursor.show()
                self._hidden_cursor = False

    def is_tty(self):
        try:
            return self.file.isatty() if self.check_tty else True
        except AttributeError:
            msg = "%s has no attribute 'isatty'. Try setting check_tty=False." % self
            raise AttributeError(msg)

    def next(self, n=1):
        now = monotonic()
        dt = now - self._ts
        self.update_avg(n, dt)
        self._ts = now
        self.index = self.index + n
        self.update()

    def iter(self, it):
        self.iter_value = None
        with self:
            for x in it:
                self.iter_value = x
                yield x
                self.next()
        del self.iter_value

    def __enter__(self):
        self.start()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.finish()


# source: https://github.com/verigak/progress/issues/58#issuecomment-471718558
def getpatchedprogress():
    # Import a clean version of the entire package.
    import progress

    # Get the current platform.
    from sys import platform

    # Check if we're on Windows.
    if platform.startswith("win"):
        # Disable HIDE_CURSOR and SHOW_CURSOR characters.
        progress.HIDE_CURSOR = ''
        progress.SHOW_CURSOR = ''
    
        # Copy over the patched Infinite class into the imported class.
        progress.Infinite = PatchedInfinite
    
    # Return the modified version of the entire package.
    return progress
